/*
 * appolomodule.c - Signature module for shinGETsu.
 * Copyright (C) 2005  shinGETsu Project.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: apollomodule.c,v 1.4 2005/03/10 05:31:21 fuktommy Exp $
 */

#include <Python.h>
#include "apollo.h"

#define SIZE 512

/*
 * Set buf[] and len[].
 */

static void
set_buf(char buf[][SIZE], int *len, int argc, char **argv)
{
    int i;
    for (i=0; i<argc; i++) {
        strncpy(buf[i], argv[i], SIZE);
        buf[i][SIZE-1] = '\0';
        len[i] = strlen(buf[i]);
        if (buf[i][len[i]-1] == '\n')
            buf[i][--len[i]] = '\0';
    }
}


/*
 * Generate key pair.
 * input: KeyGenerator
 * output: (PublicKeyStr, SecretKeyStr)
 */

static PyObject *
apollo_key_pair(PyObject *self, PyObject *args)
{
    int argc = 1;
    char *argv[argc];
    char buf[4][SIZE];
    int len[argc];

    if (!PyArg_ParseTuple(args, "s", &argv[0])) {
        return NULL;
    } else {
        set_buf(buf, len, argc, argv);
    }

    keycreate512(buf[1], buf[2], buf[0], len[0]);
    return Py_BuildValue("(s,s)", buf[1], buf[2]);
}


/*
 * Sign Message.
 * input: SignTarget, PublicKeyStr, SecretKeyStr
 * output: SignatureStr
 * 
 */

static PyObject *
apollo_sign(PyObject *self, PyObject *args)
{
    int argc = 3;
    char *argv[argc];
    char buf[4][SIZE];
    int len[argc];

    if (!PyArg_ParseTuple(args, "sss", &argv[0], &argv[1], &argv[2])) {
        return NULL;
    } else {
        set_buf(buf, len, argc, argv);
    }

    RSAsign(buf[3], buf[0], len[0], buf[1], buf[2], len[1]);
    return Py_BuildValue("s", buf[3]);
}


/*
 * Verify Signature.
 * input: SignTarget, SuspectSignature, PublicKeyStr
 * output: True / False
 */

static PyObject *
apollo_verify(PyObject *self, PyObject *args)
{
    int argc = 3;
    char *argv[argc];
    char buf[4][SIZE];
    int len[argc];
    int flag;
    
    if (!PyArg_ParseTuple(args, "sss", &argv[0], &argv[1], &argv[2])) {
        return NULL;
    } else {
        set_buf(buf, len, argc, argv);
    }

    if(RSAverify(buf[0], len[0], buf[1], buf[2], len[2]) == 0)
        flag = 1;
    else
	flag = 0;
    return Py_BuildValue("i", flag);
}


/*
 * Cut KeyStr to 11words.
 * input: KeyStr
 * output: ShortKeyStr
 */

static PyObject *
apollo_cut_key(PyObject *self, PyObject *args)
{
    int argc = 1;
    char *argv[argc];
    char buf[4][SIZE];
    int len[argc];

    if (!PyArg_ParseTuple(args, "s", &argv[0])) {
        return NULL;
    } else {
        set_buf(buf, len, argc, argv);
    }

    triphash(buf[1], buf[0], len[0]);
    return Py_BuildValue("s", buf[1]);
}


/*
 * Setup for module
 */

static PyMethodDef ApolloMethods[] = {
    {"key_pair", apollo_key_pair, METH_VARARGS, "Generate KeyPair."},
    {"sign", apollo_sign, METH_VARARGS, "Sign Message."},
    {"verify", apollo_verify, METH_VARARGS, "Verify Signature."},
    {"cut_key", apollo_cut_key, METH_VARARGS, "Cut KeyStr to 11words."},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

PyMODINIT_FUNC
initapollo(void)
{
    (void) Py_InitModule("shingetsu.apollo", ApolloMethods);
}

int
main(int argc, char *argv[])
{
    Py_SetProgramName(argv[0]);
    Py_Initialize();
    initapollo();
    return 0;
}
