#
# gateway.py - Gateway base module.
# Copyright (C) 2005  shinGETsu Project.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# $Id: gateway.py,v 1.7 2005/03/11 12:59:02 fuktommy Exp $
#

import base64
import cgi
import os.path
import re
import urllib
import sys
import time

import basecgi
import config
from cache import *
from node import *

__version__ = "$Revision: 1.7 $"


class Message(dict):

    """Multi-language message for gateway."""

    def __init__(self, file):
        dict.__init__(self)
        try:
            f = open(file)
            del_eos = re.compile(r"[\r\n]*")
            iscomment = re.compile(r"^#$").search
            for line in f:
                line = del_eos.sub("", line)
                if iscomment(line):
                    pass
                else:
                    buf = line.split("<>")
                    if len(buf) == 2:
                        buf[1] = urllib.unquote(buf[1])
                        self[buf[0]] = buf[1]
            f.close()
        except IOError:
            sys.stderr.write(file + ": IOError\n")

    def path_info(self):
        """Parse PATH_INFO and escape it."""
        path = path_info(self)
        return cgi.escape(path, True)

# Message


def search_message(accept_language):
    """Search message file.

    Example of accept_language: "ja,en-us;q=0.7,en;q=0.3"
    """
    q = {}
    lang = []
    if accept_language != "":
        for i in accept_language.split(","):
            found = re.search(r"(\S+)\s*;\s*q=(\S+)", i)
            if found:
                try:
                    lang = found.group(1)
                    q[lang] = float(found.group(2))
                except ValueError:
                    pass
            else:
                q[i] = 1

        lang = q.keys()
        lang.sort(lambda a,b: cmp(q[b], q[a]))
    lang.append(config.language)
    for i in lang:
        file = config.file_dir + "/" + "message-" + i + ".txt"
        if os.path.isfile(file):
            return Message(file)
    return None

# message


class CommonGatewayCGI(basecgi.CGI):

    root = config.root_path
    gateway_cgi = config.gateway
    list_cgi = config.list_cgi
    thread_cgi = config.thread_cgi
    note_cgi = config.note_cgi
    admin_cgi = config.admin_cgi
    message = None

    def __init__(self):
        if "HTTP_ACCEPT_LANGUAGE" in self.environ:
            al = self.environ["HTTP_ACCEPT_LANGUAGE"]
        else:
            al = ""
        self.message = search_message(al)

    def str_encode(self, query):
        """Encode for URI.

        ~ -> %7E
        """
        return urllib.quote(query)

    def str_decode(self, query):
        """Decode URI.

        %7E -> ~
        """
        return urllib.quote(query)

    def file_encode(self, type, query):
        """Encode for filename.

        ~ -> 7E
        """
        buf = [type, "_"]
        for i in query:
            buf.append("%02X" % ord(i))
        return "".join(buf)

    def file_decode(self, query):
        """Decode filename.

        7E -> ~
        """
        q = query.split("_")
        if q < 1:
            return None
        else:
            query = q[1]
        buf = []
        for i in range(0, len(query), 2):
            try:
                buf.append("%c" % int(query[i:i+2], 16))
            except ValueError:
                self.stderr.write(query + ": ValueError\n")
                return None
        return "".join(buf)

    def escape(self, msg):
        msg = re.sub(r"&", "&amp", msg)
        msg = re.sub(r"&amp;(#\d+|#[Xx][0-9A-Fa-f]+|[A-Za-z0-9]+);",
                     r"&\1",
                     msg)
        msg = re.sub(r"<", "&lt", msg)
        msg = re.sub(r">", "&gt", msg)
        msg = re.sub(r"\r", "", msg)
        msg = re.sub(r"\n", "<br>", msg)
        return msg

    def header(self, title="", rss=""):
        """Print CGI and HTTP header."""
        if rss == "":
            rss = self.gateway_cgi + "/rss"
        message = self.message
        self.stdout.write(
            "Content-Type: text/html; charset=UTF-8\n" +
            "\r\n" +
            '<?xml version="1.0" encoding="UTF-8"?>\n' +
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"\n' +
            '  "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n' +
            '<html xmlns="http://www.w3.org/1999/xhtml"' +
            ' xml:lang="' + message["lang"] + '">\n' +
            "<head>\n" +
            "<title>" + title + "</title>\n" +
            '  <link rev="made" href="http://sourceforge.net/" />\n'+
            '  <link rel="contents" href="' + self.list_cgi + '" />\n' +
            '  <link rel="stylesheet" type="text/css"' +
            ' href="' + config.css + '" />\n' +
            '  <link rel="alternate" type="application/rss+xml" title="RSS"' +
            ' href="' + rss + '" />\n' +
            "</head>\n" +
            "<body>\n" + 
            '<p class="head">\n' +
            '  <a href="' + self.gateway_cgi + '">' +
            message["top"] + "</a> |\n" +
            '  <a href="' + self.list_cgi + '">' +
            message["menu"] + "</a> |\n")

        buf = ['  <a href="' + self.gateway_cgi + '/new">' +
               message["new"] + "</a>"]
        for i in ("index", "changes", "update", "search"):
            buf.append(
                '  <a href="' + self.admin_cgi + "/" + i + '">' +
                message[i] + "</a>")
        buf.append('  <a href="' + rss + '">' + message["rss"] + "</a>")
        self.stdout.write(" |\n".join(buf))

        self.stdout.write(
              "</p>\n" +
              '<h1><a href="' + self.admin_cgi + "/search/" +
              self.str_encode(title) +
              '" class="title" title="' + message["search"] +
              '">' + title + "</a></h1>\n")

    def footer(self):
        self.stdout.write("</body>\n</html>\n")

    def localtime(self, stamp=0):
        """Return YYYY-mm-dd HH:MM."""
        return time.strftime('%Y-%m-%d %H:%M', time.localtime(int(stamp)))

    def rfc822_time(self, stamp=0):
        """Return date and time in RFC822 format."""
        return time.strftime("%a, %d %b %Y %H:%M:%S GMT",
                             time.gmtime(int(stamp)))

    def html_format(self, plain):
        buf = re.sub(r"<br>",
                     "<br />\n    ",
                     plain)
        buf = re.sub(r"(&gt;&gt;)([0-9a-f]{8})",
                     r'<a href="#r\2">\g<0></a>',
                     buf)
        buf = re.sub(r"https?://[^\x00-\x20\"'()<>\[\]\x7F-\xFF]{2,}",
                     r'<a href="\g<0>">\g<0></a>',
                     buf)
        buf = re.sub(r"\[\[<a.*?>(.*?)\]\]</a>",
                     r"[[\1]]",
                     buf)

        tmp = ""
        while len(buf) > 0:
            m = re.search(r"\[\[([^<>]+?)\]\]", buf)
            if m is not None:
                tmp += buf[:m.start()]
                tmp += self.bracket_link(m.group(1))
                buf = buf[m.end():]
            else:
                tmp += buf
                buf = ""
        return tmp

    def bracket_link(self, link):
        """Encode bracket string to link.

        See WikiWikiWeb.
        """
        m = re.search(r"^/(thread)/([^/]+)/([0-9a-f]{8})$", link)
        if m is not None:
            uri = self.thread_cgi + "/" + \
                  self.str_encode(m.group(2)) + \
                  "#r" + m.group(3)
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^/(list|thread|note)/([^/]+)$", link)
        if m is not None:
            uri = self.root + m.group(1) + ".cgi/" + \
                  self.str_encode(m.group(2))
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^([^/]+)/([0-9a-f]{8})$", link)
        if m is not None:
            uri = self.str_encode(m.group(1)) + \
                  "#r" + m.group(2)
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        m = re.search(r"^([^/]+)$", link)
        if m is not None:
            uri = self.str_encode(m.group(1))
            return '<a href="' + uri + '">[[' + link + ']]</a>'

        return "[[" + link + "]]"

    def remove_file_form(self, cache):
        filesize = str(int(cache.size/1024/1024*10)/10)
        self.stdout.write(
            '<form method="get" action="' + self.admin_cgi + '"><p>\n' +
            '  <input type="submit" value="' +
            self.message["del_file"] + '" tabindex="9" accesskey="d" />\n' +
            '  <input type="hidden" name="cmd" value="delete" />\n' +
            '  <input type="hidden" name="file" value="' +
            cache.datfile + '" />\n' +
            "  " + str(filesize) + self.message["mb"] + "\n" +
            "</p></form>\n")

    def print302(self, next):
        """Print CGI header (302 moved temporarily)."""
        self.stdout.write(
            "Content-Type: text/html; charset=ASCII\n" +
            "\n" +
            '<?xml version="1.0" encoding="ASCII"?>\n' +
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"\n' +
            '  "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">\n' +
            '<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en">\n' +
            "<head>\n" +
            "  <title>moved</title>\n" +
            '  <link rev="made" href="http://shingetsu.sourceforge.net/" />\n'+
            '  <link rel="contents" href="' + self.list_cgi + '" />\n' +
            "</head>\n" +
            "<body>\n" +
            '<p>Click and jump to <a href="' + next + '">' + next + '</a>' +
            '</p>\n' +
            '<script type="text/javascript">' +
            'window.location.href = "' + next + '";</script>\n' +
            "</body>\n</html>\n")

    def print403(self):
        """Print CGI header (403 forbidden)."""
        self.header(self.message["403"])
        self.footer()

    def print404(self, cache=None):
        """Print CGI header (404 not found)."""
        self.header(self.message["404"])
        self.stdout.write("<p>" + self.message["try_later"] + "</p>")
        if cache is not None:
            self.remove_file_form(cache)
        self.footer()

    def get_cache(self, cache):
        """Search cache from network."""
        if (cache.exists()):
            return True
        else:
            searchlist = SearchList()
            node = searchlist.search(cache)
            if node is not None:
                nodelist = NodeList()
                if node not in nodelist:
                    nodelist.append(node)
                cache.get_range(node)
                return True
            else:
                return False

    def print_new_element_form(self, parent=None):
        if parent is not None:
            datfile = parent.datfile
            cginame = self.list_cgi
        else:
            datfile = ""
            cginame = self.gateway_cgi

        self.stdout.write(
            '<form method="get" action="' + cginame + '"><p>\n' +
            '  <input type="hidden" name="cmd" value="new" />\n' +
            '  <input type="hidden" name="file" value="' + datfile + '" />\n' +
            '  <input type="submit" value="' + self.message["new"] + '"' +
            ' name="submit" tabindex="1" accesskey="w" />\n')

        for type in config.types:
            self.stdout.write(
                '  <input type="radio" value="' + type + '" name="type"' +
                ' tabindex="2" accesskey="t" />' +
                self.message[type] + "\n")

        if parent is not None:
            self.stdout.write(
                '  <input type="checkbox" name="dopost" value="dopost"' +
                ' checked="checked" tabindex="2" accesskey="s" />' +
                self.message["send"] + "\n")

        self.stdout.write(
            "  <br />\n" +
            self.message["name"] + ": " +
            '<input name="name" size="19" value=""' +
            ' tabindex="3" accesskey="t" />\n' +
            "</p></form>\n")

    def error_time(self):
        from math import sqrt, log, sin, pi
        from random import random
        (r1, r2) = (random(), random())
        normal = sqrt(-2 * log(r1)) * sin(2 * pi * r2)
        now = int(time.time())
        error = int(config.time_error * normal)
        return now + error

    def do_post(self, path, form):
        """Post article."""
        try:
            attach = form["attach"]
        except KeyError:
            attach = None
            str_attach = ""

        if attach and attach.file:
            if (len(attach.value) > config.file_limit*1024*1024):
                self.header(self.message["big_file"])
                self.footer()
                return
            str_attach = base64.encodestring(attach.value)
        guess_suffix = "txt"
        if attach and attach.filename != "":
            found = re.search(r"\.([^.]+)$", attach.filename)
            if found:
                guess_suffix = found.group(1).lower()

        suffix = form.getfirst("suffix", "")
        if (suffix == "") or (suffix == "AUTO"):
            suffix = guess_suffix
        elif suffix.startswith("."):
            suffix = suffix[1:].lower()
        else:
            suffix = suffix.lower()

        if form.getfirst("error", "") != "":
            stamp = self.error_time()
        else:
            stamp = int(time.time())

        body = {}
        for key in ("base_stamp", "base_id","name", "mail", "body"):
            if form.getfirst(key, "") != "":
                body[key] = self.escape(form.getfirst(key, ""))
        if suffix != "":
            body["suffix"] = self.escape(suffix)
        if str_attach != "":
            body["attach"] = str_attach

        if len(body) == 0:
            self.header(self.message["null_article"])
            self.footer()
            return

        cache = Cache(form.getfirst("file"))
        rec = Record(datfile=cache.datfile)
        passwd = form.getfirst("passwd", "")
        id = rec.build(stamp, body, passwd=passwd)
        cache.add_data([rec])

        if form.getfirst("dopost", "") != "":
            updatelist = UpdateList()
            updatelist.append(rec)
            updatelist.sync()
            nodelist = NodeList()
            nodelist.tell_update(datfile=cache.datfile, stamp=stamp, id=id)

        return id[:8]

    def do_new_element(self, path, form):
        """Make new list element."""
        stamp = self.error_time()

        if form.getfirst("name", "") == "":
            self.header(self.message["null_name"])
            self.footer()
            return
        elif re.search(r"/|\]\]", form.getfirst("name", "")):
            self.header(self.message["bad_name"])
            self.footer()
            return
        elif form.getfirst("type", "") == "":
            self.header(self.message["null_type"])
            self.footer()
            return

        cache = Cache(form.getfirst("file"))
        rec = Record(datfile=cache.datfile)
        id = rec.build(stamp, {"type": form.getfirst("type", ""),
                               "link": form.getfirst("name", "")})
        cache.add_data([rec])

        if form.getfirst("dopost", "") != "":
            updatelist = UpdateList()
            updatelist.append(rec)
            updatelist.sync()
            nodelist = NodeList()
            nodelist.tell_update(datfile=cache.datfile, stamp=stamp, id=id)

        return True